/*----------------------------------------------------------------------------
 * libimgloader - Loading images made simple
 * Copyright (C) 2008 Arne Bochem <libimgloader@arneb.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *----------------------------------------------------------------------------*/

#include "imgloader_common.h"
#include "bmp_common.h"
#include "read_bmp.h"
#include <math.h>

/* Code */

frames_t *decode_palette_bmp(FILE *fh, int w, int h, int bitmap_offset, int bitmap_size, int bpp, int colors_used, int entry_size, int compression)
{
	frames_t *result;
	uint8_t *palette;
	uint8_t *bitmap;
	int palette_size;
	int row_size;
	int top_first = 0;
	int ppb = 8 / bpp; /* pixels per byte */
	int shift;
	int idx;
	int i, j;
	int y;

	/* In Windows V3 bitmaps, negative height means, that the bitmap is top-to-bottom instead of bottom-to-top */
	if (h < 0)
	{
		top_first = 1;
		h = -h;
		y = 0;
	}
	else
		y = h - 1;

	/* Let's be lazy for now */
	if (compression != BMP_COMP_RGB)
	{
		fprintf(stderr, "Compressed BMP not yet supported.\n");
		return NULL;
	}

	/* Pictures with more than 256 colors don't use palette */
	if (colors_used > 256)
	{
		fprintf(stderr, "Invalid BMP file.\n");
		return NULL;
	}

	/* Reduced number of palette colors, or just the full 2^n colors? */
	if (colors_used)
		palette_size = colors_used;
	else
		palette_size = (int)pow(2, bpp);

	/* Number of bytes in a row is width / pixels per byte */
	row_size = w / ppb;

	/* Rows are padded to multiples of 4B */
	if (row_size % 4)
		row_size += 4 - (row_size % 4);

	/* If we don't have a bitmap size specified, we'll just assume it to be row_size * height */
	if (!bitmap_size)
		bitmap_size = row_size * h;
	
	/* Allocate space for palette, raw bitmap data, and result image */
	FATALLOC(palette, palette_size * entry_size, "palette");
	FATALLOC(bitmap, bitmap_size, "bitmap data");
	result = alloc_images(1, w, h);

	/* Read palette */
	fread(palette, entry_size, palette_size, fh);

	/* Seek to start of bitmap data */
	fseek(fh, bitmap_offset, SEEK_SET);

	/* And read it */
	fread(bitmap, 1, bitmap_size, fh);

	for (j = 0; j < h; j++)
	{
		for (i = 0; i < w; i++)
		{
			/* Funky bit magic (FIXME: write better comment) */
			shift = (ppb - 1 - (i % ppb)) * bpp;
			idx = (bitmap[i / ppb] & (bmp_bpp_masks[bpp] << shift)) >> shift;

			/* Let's be lenient and just skip out-of-bound indices */
			if (idx >= palette_size)
				continue;

			result->frames[0][(y * w + i) * 3 + 0] = palette[entry_size * idx + 2];
			result->frames[0][(y * w + i) * 3 + 1] = palette[entry_size * idx + 1];
			result->frames[0][(y * w + i) * 3 + 2] = palette[entry_size * idx + 0];
		}

		/* Process next row */
		bitmap += row_size;

		/* Adjust ine in target image in correct direction*/
		if (top_first)
			y++;
		else
			y--;
	}

	bitmap -= row_size * h;

	/* Free temporary data structures */
	free(bitmap);
	free(palette);

	return result;
}

frames_t *decode_rgb_bmp(FILE *fh, int w, int h, int bitmap_offset, int bitmap_size, int pixel_size, int compression)
{
	frames_t *result;
	uint8_t *bitmap;
	int top_first = 0;
	int row_size;
	int i, j;
	int y;

	/* In Windows V3 bitmaps, negative height means, that the bitmap is top-to-bottom instead of bottom-to-top */
	if (h < 0)
	{
		top_first = 1;
		h = -h;
		y = 0;
	}
	else
		y = h - 1;

	/* Let's be lazy for now */
	if (compression != BMP_COMP_RGB)
	{
		fprintf(stderr, "Compressed BMP not yet supported.\n");
		return NULL;
	}

	/* Make sure we don't have pixels beyond the actual data */
	if (bitmap_size < w * h * pixel_size)
	{
		/* Wait, maybe somebody just forgot to set the value? */
		if (!bitmap_size)
			bitmap_size = w * h * pixel_size;
		else
		{
			fprintf(stderr, "Invalid BMP data length.\n");
			return NULL;
		}
	}

	/* Allocate space for raw bitmap data and result */
	FATALLOC(bitmap, bitmap_size, "bitmap data");
	result = alloc_images(1, w, h);

	/* Seek to start of bitmap data */
	fseek(fh, bitmap_offset, SEEK_SET);

	/* And read it */
	fread(bitmap, 1, bitmap_size, fh);

	/* Row size is width * bytes per pixel */
	row_size = w * pixel_size;

	/* Rows are padded to multiples of 4B */
	if (row_size % 4)
		row_size += 4 - (row_size % 4);

	for (j = 0; j < h; j++)
	{
		for (i = 0; i < w; i++)
		{
			result->frames[0][(y * w + i) * 3 + 0] = bitmap[j * row_size + i * pixel_size + 2];
			result->frames[0][(y * w + i) * 3 + 1] = bitmap[j * row_size + i * pixel_size + 1];
			result->frames[0][(y * w + i) * 3 + 2] = bitmap[j * row_size + i * pixel_size + 0];
		}

		/* Adjust ine in target image in correct direction*/
		if (top_first)
			y++;
		else
			y--;
	}

	free(bitmap);

	return result;
}

frames_t *decode_os2_v1_bmp(FILE *fh, struct bmp_file_header file_header, int *w, int *h)
{
	struct bmp_info_header_os2_v1 info_header;

	if (fread(&info_header, sizeof(struct bmp_info_header_os2_v1), 1, fh) < 1)
	{
		fprintf(stderr, "Could not read BMP info header.\n");
		return NULL;
	}

	if (info_header.planes != 1)
	{
		fprintf(stderr, "Invalid BMP info header.\n");
		return NULL;
	}

	*w = info_header.width;
	*h = info_header.height;

	switch (info_header.depth)
	{
		case 1: case 4: case 8:
			return decode_palette_bmp(fh, *w, *h, file_header.bmp_offset, 0, info_header.depth, 0, 3, BMP_COMP_RGB);

		case 24:
			return decode_rgb_bmp(fh, *w, *h, file_header.bmp_offset, 0, 3, BMP_COMP_RGB);

		default:
			fprintf(stderr, "Invalid color depth for OS/2 V1 bitmap.\n");
			return NULL;
	}

	return NULL; /* should never be reached, but let's keep it */
}

frames_t *decode_win_v3_bmp(FILE *fh, struct bmp_file_header file_header, int *w, int *h)
{
	struct bmp_info_header_win_v3 info_header;

	if (fread(&info_header, sizeof(struct bmp_info_header_win_v3), 1, fh) < 1)
	{
		fprintf(stderr, "Could not read BMP info header.\n");
		return NULL;
	}

	if (info_header.planes != 1)
	{
		fprintf(stderr, "Invalid BMP info header.\n");
		return NULL;
	}

	*w = info_header.width;
	*h = info_header.height;

	switch (info_header.depth)
	{
		case 1: case 4: case 8:
			return decode_palette_bmp(fh, *w, *h, file_header.bmp_offset, info_header.image_size, info_header.depth, info_header.colors_used, 4, info_header.compression);

		case 16:
			fprintf(stderr, "16bit BMP decoding is not supported yet.\n");
			return NULL;

		case 24:
			return decode_rgb_bmp(fh, *w, *h, file_header.bmp_offset, info_header.image_size, 3, info_header.compression);

		case 32:
			return decode_rgb_bmp(fh, *w, *h, file_header.bmp_offset, info_header.image_size, 4, info_header.compression);

		default:
			fprintf(stderr, "Invalid color depth for Windows V3 bitmap.\n", info_header.depth);
			return NULL;
	}

	return NULL;
}

frames_t *read_bmp(char *filename, int *w, int *h)
{
	struct bmp_file_header file_header;
	uint32_t info_header_length;
	frames_t *result = NULL;
	FILE *fh;

	/* Open file handle */
	if ((fh = fopen(filename, "rb")) == NULL)
	{
		perror("Cannot open BMP file for reading");
		return NULL;
	}

	if (fread(&file_header, sizeof(struct bmp_file_header), 1, fh) < 1)
	{
		fprintf(stderr, "Could not read BMP file header.\n");
		return NULL;
	}

	if (file_header.type != 0x4d42 || file_header.reserved_1 + file_header.reserved_2 != 0) /* BM....0000.... */
	{
		fprintf(stderr, "Invalid BMP file header.\n");
		return NULL;
	}

	if (fread(&info_header_length, sizeof(uint32_t), 1, fh) < 1)
	{
		fprintf(stderr, "Could not read BMP info header.\n");
		return NULL;
	}

	switch (info_header_length)
	{
		case 12: /* length of OS/2 V1 info header */
			result = decode_os2_v1_bmp(fh, file_header, w, h);
			break;

		case 40: /* length of Windows V3 info header */
			result = decode_win_v3_bmp(fh, file_header, w, h);
			break;
	}

	/* Close file handle */
	fclose(fh);

	return(result);
}
