/*----------------------------------------------------------------------------
 * libimgloader - Loading images made simple
 * Copyright (C) 2008 Arne Bochem <libimgloader@arneb.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *----------------------------------------------------------------------------*/

#include "imgloader_common.h"
#include <gif_lib.h>
#include "read_gif.h"

/* Code */

/* -1 = failure, 0 = unspecified (do nothing), 1 = leave in place, 2 = restore background, 3 = restore to previous */
int make_image(char *img, GifFileType *gif, int id, int *delay)
{
	ExtensionBlock *exts = gif->SavedImages[id].ExtensionBlocks;
	uint8_t *pixels = gif->SavedImages[id].RasterBits;
	ColorMapObject *colormap;
	int interlaced_starts[4] = {0, 4, 2, 1};
	int interlaced_steps[4] = {8, 8, 4, 2};
	int transparent = -1;
	int disposal = 0;
	int full_w = gif->SWidth;
	int l = gif->SavedImages[id].ImageDesc.Left;
	int t = gif->SavedImages[id].ImageDesc.Top;
	int w = gif->SavedImages[id].ImageDesc.Width;
	int h = gif->SavedImages[id].ImageDesc.Height;
	uint8_t idx;
	int i, j, k;

	/* Make sure nothing goes out-of-bounds */
	if (w + l > gif->SWidth || h + t > gif->SHeight)
		return -1;
	w += l;
	h += t;

	/* Determine color map */
	if (gif->SavedImages[id].ImageDesc.ColorMap != NULL)
		colormap = gif->SavedImages[id].ImageDesc.ColorMap;
	else
		colormap = gif->SColorMap;

	if (colormap == NULL)
		return -1;

	/* Get transparency and frame display time */
	for (i = 0; i < gif->SavedImages[id].ExtensionBlockCount; i++)
	{
		if (exts[i].Function == GRAPHICS_EXT_FUNC_CODE && exts[i].ByteCount == 4)
		{
			*delay = *(uint16_t*)(exts[i].Bytes + 1) * 10;

			if (exts[i].Bytes[0] & 1)
				transparent = (uint8_t)(exts[i].Bytes[3]);
			else
				transparent = -1;

			disposal = (exts[i].Bytes[0] & 28) >> 2;
		}
	}

	if (!gif->SavedImages[id].ImageDesc.Interlace)
		for (j = t; j < h; j++)
		{
			for (i = l; i < w; i++)
			{
				/* Get color index for next pixel */
				idx = *(pixels++);

				/* Make sure it's not out of bounds */
				if (idx >= colormap->ColorCount)
					return -1;

				/* Skip drawing it, if it is transparent */
				if (transparent > -1 && idx == transparent)
					continue;

				/* Look up index in color map and set pixel */
				img[(j * full_w + i) * 3 + 0] = colormap->Colors[idx].Red;
				img[(j * full_w + i) * 3 + 1] = colormap->Colors[idx].Green;
				img[(j * full_w + i) * 3 + 2] = colormap->Colors[idx].Blue;
			}
		}
	else
		for (k = 0; k < 4; k++) /* Interlaced GIF images have reordered lines. */ 
			/* The first 1/8 lines are every 8th, second 1/8 are every 8th+4, next 1/4 of lines are every 4th+2, next 1/2 of lines are every 2nd line. Confused yet? */
			for (j = interlaced_starts[k] + t; j < h; j += interlaced_steps[k])
			{
				/* Same as above */
				for (i = l; i < w; i++)
				{
					idx = *(pixels++);

					if (idx >= colormap->ColorCount)
						return -1;

					if (transparent > -1 && idx == transparent)
						continue;

					img[(j * full_w + i) * 3 + 0] = colormap->Colors[idx].Red;
					img[(j * full_w + i) * 3 + 1] = colormap->Colors[idx].Green;
					img[(j * full_w + i) * 3 + 2] = colormap->Colors[idx].Blue;
				}
			}

	return disposal;
}

/* -1 = failure, 0 = success */
int fill_with_bgcolor(char *img, GifFileType *gif)
{
	int i;

	if (gif->SColorMap == NULL)
		return -1;

	for (i = 0; i < gif->SWidth * gif->SHeight * 3; i += 3)
	{
		img[i + 0] = gif->SColorMap->Colors[gif->SBackGroundColor].Red;
		img[i + 1] = gif->SColorMap->Colors[gif->SBackGroundColor].Green;
		img[i + 2] = gif->SColorMap->Colors[gif->SBackGroundColor].Blue;
	}

	return 0;
}

frames_t *read_gif(char *filename, int *w, int *h)
{
	frames_t *result;
	GifFileType *gif;
	int count;
	int delay = 0;
	int disp;
	int i;

	if ((gif = DGifOpenFileName(filename)) == NULL)
		return NULL;

	if (DGifSlurp(gif) != GIF_OK)
	{
		DGifCloseFile(gif);
		return NULL;
	}

	*w = gif->SWidth;
	*h = gif->SHeight;
	count = gif->ImageCount;

	result = alloc_images(count, *w, *h);

	if (fill_with_bgcolor(result->frames[0], gif) == -1)
		return NULL;

	for (i = 0; i < count; i++)
	{
		if ((disp = make_image(result->frames[i], gif, i, &delay)) == -1)
			return NULL;

		if (delay != 0) /* use default of 100ms for insane speeds (?) */
			result->delays[i] = delay;

		if (i + 1 < count)
		{
			switch (disp)
			{
				case 0:
					memset(result->frames[i + 1], 0, *w * *h * 3); /* let's be nice and give a clean field of zeros */
					break;

				case 1:
					memcpy(result->frames[i + 1], result->frames[i], *w * *h * 3); /* copy current frame to next */
					break;

				case 2:
					fill_with_bgcolor(result->frames[i + 1], gif); /* set next frame to bgcolor */
					break;

				case 3:
					if (i > 0)
						memcpy(result->frames[i + 1], result->frames[i - 1], *w * *h * 3); /* copy previous frame to next */
					break;
			}
		}
	}

	DGifCloseFile(gif);

	return result;
}

/*int main (int argc, char *argv[])
{
	read_gif(argv[1], NULL, NULL);
}*/
