/*----------------------------------------------------------------------------
 * libimgloader - Loading images made simple
 * Copyright (C) 2008 Arne Bochem <libimgloader@arneb.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *----------------------------------------------------------------------------*/

#include "imgloader_common.h"
#include <png.h>
#include "read_png.h"

/* Code */

frames_t *read_png(char *filename, int *w, int *h)
{
	frames_t *result;
	FILE *fh;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep *row_pointers;
	uint8_t header[8], *image;
	int read_bytes, bit_depth, color_type, i;

	if ((fh = fopen(filename, "rb")) == NULL)
	{
		perror("Cannot open PNG file for reading");
		return NULL;
	}

	/* Check header */
	read_bytes = fread(header, 1, 8, fh);
	if (png_sig_cmp(header, 0, read_bytes))
	{
		fprintf(stderr, "Image is not a PNG file: %s\n", filename);
		return NULL;
	}

	/* Initialize png struct */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL)
	{
		fprintf(stderr, "Cannot create png_ptr.\n");
		return NULL;
	}

	/* Initialize info struct */
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fprintf(stderr, "Cannot create info_ptr.\n");
		return NULL;
	}

	/* Set long jump stuff (weird..?) */
	if (setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(fh);
		fprintf(stderr, "Error while reading PNG file: %s\n", filename);
		return NULL;
	}

	/* Initialize IO */
	png_init_io(png_ptr, fh);

	/* Set signature bytes */
	png_set_sig_bytes(png_ptr, read_bytes);

	/* Read file info */
	png_read_info(png_ptr, info_ptr);

	/* Gather information */
	*w = png_get_image_width(png_ptr, info_ptr);
	*h = png_get_image_height(png_ptr, info_ptr);
	bit_depth = png_get_bit_depth(png_ptr, info_ptr);
	color_type = png_get_color_type(png_ptr, info_ptr);

	/* Fix image format */
	if (color_type == PNG_COLOR_TYPE_PALETTE)
	{
		png_set_palette_to_rgb(png_ptr);
		png_set_strip_alpha(png_ptr);
	}

	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
	{
		png_set_gray_1_2_4_to_8(png_ptr);
	}

	if (bit_depth == 16)
	{
		png_set_strip_16(png_ptr);
	}

	if (color_type & PNG_COLOR_MASK_ALPHA)
	{
		png_set_strip_alpha(png_ptr);
	}

	if (bit_depth < 8)
	{
		png_set_packing(png_ptr);
	}

	if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
	{
		png_set_gray_to_rgb(png_ptr);
	}

	/* Update image information */
	png_read_update_info(png_ptr, info_ptr);

	/* Allocate image memory */
	result = alloc_images(1, *w, *h);
	image = result->frames[0];

	/* Allocate row pointer memory */
	FATALLOC(row_pointers, *h * sizeof(png_bytep), "row pointers");

	/* Set row pointers */
	for (i = 0; i < *h; i++)
	{
		row_pointers[i] = image + i * *w * 3;
	}

	/* Read image */
	png_read_image(png_ptr, row_pointers);

	/* End read */
	png_read_end(png_ptr, NULL);

	/* Free memory */
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	free(row_pointers);

	/* Close file handle */
	fclose(fh);

	return(result);
}
