/*----------------------------------------------------------------------------
 * libimgloader - Loading images made simple
 * Copyright (C) 2008 Arne Bochem <libimgloader@arneb.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *----------------------------------------------------------------------------*/

#include "read_targa.h"
#include "targa_common.h"

/* Code */

int targa_decode_rgb(char *filename, uint8_t *image, struct targa_file tga, uint8_t *color_map, int rle)
{
	uint8_t *im = tga.image_data;
	int pixel_count = 0;
	int pixel_total = tga.width * tga.height;
	int base, skip, i, count, is_rle, type_det;
	int x = 0;
	int x_mod = 1;
	int x_bound = tga.width;
	int x_init = 0;
	int y = tga.height - 1;
	int y_mod = -1;
	int y_base;

	/* Check vertical direction */
	if (TARGA_LINE_ORIGIN(tga.image_descriptor))
	{
		y = 0;
		y_mod = 1;
	}

	/* Check horizontal direction */
	if (TARGA_COL_ORIGIN(tga.image_descriptor))
	{
		x = tga.width - 1;
		x_init = x;
		x_mod = -1;
		x_bound = -1;
	}

	/* Set pixel size */
	skip = tga.pixel_size / 8;
	if ((skip < 1) || (skip > 4))
	{
		fprintf(stderr, "TGA %s has invalid pixel size %u.\n", filename, tga.pixel_size);
		return 0;
	}

	/* Set pixel type */
	type_det = skip;
	if (color_map != NULL)
	{
		/* Color mapped pixels have negative types */
		type_det = type_det * -1;
	}

	/* Initial y base */
	y_base = y * tga.width;
	while (pixel_count < pixel_total)
	{
		base = (y_base + x) * 3;

		/* Check for RLE encoded image */
		if (rle)
		{
			/* Determine whether this is an RLE or a raw packet */
			is_rle = TARGA_IS_RLE(im[0]);

			/* Get count of repetitions or contained pixels */
			count = TARGA_PACK_LEN(im[0]);

			/* Step over RLE packet header */
			im++;
		}
		else
		{
			/* No RLE at all */
			count = 1;
			is_rle = 0;
		}

		/* Loops for RLE packets or for whole picture */
		for (i = 0; (i < count) && (pixel_count < pixel_total); i++)
		{
			/* Set pixel */
			switch (type_det)
			{
				/* Color mapped, 16bit index */
				case -2:
					image[base] = color_map[(uint16_t)*im * 3];
					image[base + 1] = color_map[(uint16_t)*im * 3 + 1];
					image[base + 2] = color_map[(uint16_t)*im * 3 + 2];
					break;

				/* Color mapped, 8bit index */
				case -1:
					image[base] = color_map[im[0] * 3];
					image[base + 1] = color_map[im[0] * 3 + 1];
					image[base + 2] = color_map[im[0] * 3 + 2];
					break;

				/* Grayscale */
				case 1:
					image[base] = im[0];
					image[base + 1] = im[0];
					image[base + 2] = im[0];
					break;

				/* 16bit RGB (BGR) */
				case 2:
					image[base] = TARGA_16_RED((uint16_t)*im);
					image[base + 1] = TARGA_16_GREEN((uint16_t)*im);
					image[base + 2] = TARGA_16_BLUE((uint16_t)*im);
					break;

				/* 24bit and 32bit RGB (BGR) */
				case 3:
				case 4:
					image[base] = im[2];
					image[base + 1] = im[1];
					image[base + 2] = im[0];
					break;
			}

			/* Count pixels */
			pixel_count++;

			/* Prepare next pixel in RLE mode */
			base += 3;

			/* Set position for next pixel */
			x += x_mod;
			if (x == x_bound)
			{
				x = x_init;
				y += y_mod;
				y_base = y * tga.width;
			}

			/* Step forward for raw pixels */
			if (!is_rle)
			{
				im += skip;
			}
		}

		/* Step forward for RLE encoded packets */
		if (is_rle)
		{
			im += skip;
		}
	}

	return 1;
}

uint8_t *targa_make_colormap (struct targa_file tga)
{
	int skip = (tga.color_map_entry_size + 1) / 8; /* 15bit is the same as 16bit */
	uint8_t *color_map, *in;
	int i;

	/* Allocate memory for color map */
	FATALLOC(color_map, 3 * tga.color_map_length, "color map");
	in = tga.color_map;

	/* Kinda like image parsing, just with less cases */
	for (i = 0; i < tga.color_map_length; i++)
	{
		switch (skip)
		{
			/* 15bit and 16bit color maps */
			case 2:
				color_map[i * 3] = TARGA_16_RED((uint16_t)*in);
				color_map[i * 3 + 1] = TARGA_16_GREEN((uint16_t)*in);
				color_map[i * 3 + 2] = TARGA_16_BLUE((uint16_t)*in);
				break;

			/* 24bit and 32 bit color maps */
			case 3:
			case 4:
				color_map[i * 3] = in[2];
				color_map[i * 3 + 1] = in[1];
				color_map[i * 3 + 2] = in[0];
				break;
		}

		/* Step forward */
		in += skip;
	}

	/* Offset color map if offset is specified */
	color_map -= tga.color_map_origin * 3;

	return(color_map);
}

frames_t *read_targa(char *filename, int *w, int *h)
{
	FILE *fh;
	frames_t *result;
	uint8_t *image;
	uint8_t *color_map = NULL;
	struct targa_file tga;
	int length;

	/* Open file */
	if ((fh = fopen(filename, "rb")) == NULL)
	{
		perror("Cannot open TGA file for reading");
		return NULL;
	}

	/* Get file length */
	fseek(fh, 0, SEEK_END);
	length = ftell(fh) - sizeof(struct targa_file) + sizeof(uint8_t *) * 3;
	rewind(fh);

	/* Read header */
	fread(&tga, sizeof(struct targa_file) - sizeof(uint8_t *) * 3, 1, fh);

	/* Set outside width and height */
	*w = tga.width;
	*h = tga.height;

	/* Check for substance */
	if (!(tga.width && tga.height))
	{
		fprintf(stderr, "TGA %s has zero size.\n", filename);
		return NULL;
	}

	/* Allocate and read image ID */
	if (tga.id_field_len)
	{
		FATALLOC(tga.image_id, tga.id_field_len, "image ID field");
		fread(tga.image_id, tga.id_field_len, 1, fh);
		length -= tga.id_field_len;
	}

	/* Allocate and read color map, if applicable */
	if (tga.color_map_type)
	{
		FATALLOC(tga.color_map, tga.color_map_length * tga.color_map_entry_size / 8, "color map");
		fread(tga.color_map, tga.color_map_length * tga.color_map_entry_size / 8, 1, fh);
		length -= tga.color_map_length * tga.color_map_entry_size / 8;
	}
	else
	{
		if ((tga.frames_type == 1) || (tga.frames_type == 9))
		{
			fprintf(stderr, "TGA color map in %s not found.\n", filename);
			return NULL;
		}
	}

	/* Allocate and read image data */
	if (length)
	{
		FATALLOC(tga.image_data, length, "image data");
		fread(tga.image_data, length, 1, fh);
	}
	else
	{
		fprintf(stderr, "TGA %s contains no image data.\n", filename);
		return NULL;
	}

	/* Set image pointer */
	result = alloc_images(1, tga.width, tga.height);
	image = result->frames[0];

	/* Choose decoding mechanisms */
	switch (tga.frames_type)
	{
		case 1:
			color_map = targa_make_colormap(tga);
		case 2:
		case 3:
			if (!targa_decode_rgb(filename, image, tga, color_map, 0))
				return NULL;
			break;

		case 9:
			color_map = targa_make_colormap(tga);
		case 10:
		case 11:
			if (!targa_decode_rgb(filename, image, tga, color_map, 1))
				return NULL;
			break;

		default:
			fprintf(stderr, "Type %u in %s is not a supported TGA format.\n", tga.frames_type, filename);
			return NULL;
			break;
	}

	/* Free memory */
	if (tga.id_field_len)
	{
		free(tga.image_id);
	}

	if (tga.color_map && (tga.color_map_length * tga.color_map_entry_size / 8))
	{
		free(tga.color_map);
	}

	if (color_map != NULL)
	{
		free(color_map);
	}

	free(tga.image_data);


	return(result);
}
